@@ -23,14 +23,15 @@ class _Projection(utils.Projection):
 
 
 class ObliqueProjection(_Projection):
-    def __init__(self, linear_system, U, **kwargs):
+    def __init__(self, linear_system, U, qr_reorthos=0, **kwargs):
         '''Oblique projection for left deflation.'''
         # preprocess and store input
         self.linear_system = linear_system
         (N, d) = U.shape
 
         # orthogonalize U in the Minv-inner-product
-        U, _ = utils.qr(U, ip_B=linear_system.get_ip_Minv_B())
+        U, _ = utils.qr(U, ip_B=linear_system.get_ip_Minv_B(),
+                        reorthos=qr_reorthos)
 
         self.U = U
         '''An orthonormalized basis of the deflation space ``U`` with respect
@@ -41,6 +42,8 @@ class ObliqueProjection(_Projection):
         '''Result of application of operator to deflation space, i.e.,
         :math:`M_lAM_rU`.'''
 
+        self._MAU = None
+
         # call Projection constructor
         super(_Projection, self).__init__(self.AU, self.U,
                                           ip_B=linear_system.ip_B,
@@ -59,6 +62,14 @@ class ObliqueProjection(_Projection):
             c = self.WR.dot(scipy.linalg.solve_triangular(self.VR, c))
         return z + self.W.dot(c)
 
+    @property
+    def MAU(self):
+        '''Result of preconditioned operator to deflation space, i.e.,
+        :math:`MM_lAM_rU`.'''
+        if self._MAU is None:
+            self._MAU = self.linear_system.M * self.AU
+        return self._MAU
+
 
 class _DeflationMixin(object):
     '''Mixin class for deflation in Krylov subspace methods.
@@ -73,13 +84,19 @@ class _DeflationMixin(object):
     All other parameters are passed through to the underlying solver from
     :py:mod:`~krypy.linsys`.
     '''
-    def __init__(self, linear_system, U=None, *args, **kwargs):
+    def __init__(self, linear_system, U=None,
+                 projection_kwargs=None, *args, **kwargs):
         if U is None:
             U = numpy.zeros((linear_system.N, 0))
-        projection = ObliqueProjection(linear_system, U)
+        if projection_kwargs is None:
+            projection_kwargs = {}
+
+        # construct projection
+        projection = ObliqueProjection(linear_system, U, **projection_kwargs)
         self.projection = projection
         '''Projection that is used for deflation.'''
 
+        # retrieve E=ip_B(U,AU) from projection
         if projection.Q is None and projection.R is None:
             E = numpy.eye(U.shape[1])
         else:
@@ -99,6 +116,7 @@ class _DeflationMixin(object):
         self._B_ = None
 
         super(_DeflationMixin, self).__init__(linear_system,
+                                              dtype=U.dtype,
                                               *args, **kwargs)
 
     def _solve(self):
@@ -161,6 +179,46 @@ class _DeflationMixin(object):
                                        ip_B=ls.ip_B)
         return self._B_
 
+    def estimate_time(self, nsteps, ndefl, deflweight=1.0):
+        '''Estimate time needed to run nsteps iterations with deflation
+
+        Uses timings from :py:attr:`linear_system` if it is an instance of
+        :py:class:`~krypy.linsys.TimedLinearSystem`. Otherwise, an
+        :py:class:`~krypy.utils.OtherError`
+        is raised.
+
+        :param nsteps: number of iterations.
+        :param ndefl: number of deflation vectors.
+        :param deflweight: (optional) the time for the setup and application of
+          the projection for deflation is multiplied by this factor. This can
+          be used as a counter measure for the evaluation of Ritz vectors.
+          Defaults to 1.
+        '''
+        # get ops for nsteps of this solver
+        solver_ops = self.operations(nsteps)
+
+        # define ops for deflation setup + application with ndefl deflation
+        # vectors
+        proj_ops = {'A': ndefl,
+                    'M': ndefl,
+                    'Ml': ndefl,
+                    'Mr': ndefl,
+                    'ip_B': (ndefl*(ndefl+1)/2
+                             + ndefl**2 + 2*ndefl*solver_ops['Ml']),
+                    'axpy': (ndefl*(ndefl+1)/2 + ndefl*ndefl
+                             + (2*ndefl+2)*solver_ops['Ml'])
+                    }
+
+        # get timings from linear_system
+        if not isinstance(self.linear_system, linsys.TimedLinearSystem):
+            raise utils.RuntimeError(
+                'A `TimedLinearSystem` has to be used in order to obtain '
+                'timings.')
+        timings = self.linear_system.timings
+
+        return (timings.get_ops(solver_ops)
+                + deflweight*timings.get_ops(proj_ops))
+
 
 class DeflatedCg(_DeflationMixin, linsys.Cg):
     '''Deflated preconditioned CG method.
@@ -223,6 +281,8 @@ class Arnoldifyer(object):
         V = deflated_solver.V
         U = deflated_solver.projection.U
         AU = deflated_solver.projection.AU
+        ls = deflated_solver.linear_system
+        MAU = deflated_solver.projection.MAU
 
         # get dimensions
         n_, n = self.n_, self.n = H.shape
@@ -241,15 +301,24 @@ class Arnoldifyer(object):
         self.A_norm = numpy.linalg.norm(self.M, 2)
 
         if d > 0:
-            # rank-revealing QR decomp of projected AU
-            Q, R, P = scipy.linalg.qr(AU - U.dot(E) - V.dot(B_),
+            # rank-revealing QR decomp of projected MAU
+            Q, R, P = scipy.linalg.qr(MAU - U.dot(E) - V.dot(B_),
                                       mode='economic', pivoting=True)
+
+            # inverse permutation
             P_inv = numpy.argsort(P)
 
             # rank of R
             l = (numpy.abs(numpy.diag(R)) > 1e-14*self.A_norm).sum()
             Q1 = Q[:, :l]
             self.R12 = R[:l, P_inv]
+
+            # reorthonormalize in correct inner product.
+            # the above can be removed if the rank-revealing factorization
+            # is carried out directly in the appropriate inner product
+            Q1, Rt = utils.qr(Q1, ip_B=ls.get_ip_Minv_B())
+            self.R12 = Rt.dot(self.R12)
+
             # residual helper matrix
             self.N = numpy.c_[numpy.eye(l+n_-n, n_-n),
                               numpy.r_[B_[n:, :],
@@ -296,9 +365,10 @@ class Arnoldifyer(object):
         d = self.d
         k = Wt.shape[1]
 
-        # get orthonormal basis of Wt^\perp
+        # get orthonormal basis of Wt and Wt^\perp
         if k > 0:
             Wto, _ = scipy.linalg.qr(Wt)
+            Wt = Wto[:, :k]
             Wto = Wto[:, k:]
         else:
             Wto = numpy.eye(Wt.shape[0])
@@ -327,8 +397,13 @@ class Arnoldifyer(object):
 
         # Arnoldify
         WtoQ = Q.apply(Wto.T.conj()).T.conj()
+
+        # TODO: replace with scipy's Hessenberg when bug is fixed
+        #from scipy.linalg import hessenberg
+        #Hh, T = hessenberg(
         Hh, T = utils.hessenberg(
-            Q.apply(Wto.T.conj().dot(self.J).dot(Pt*(self.L.dot(WtoQ)))))
+            Q.apply(Wto.T.conj().dot(self.J).dot(Pt*(self.L.dot(WtoQ)))),
+            calc_q=True)
 
         QT = Q.apply(T)
 
@@ -357,22 +432,33 @@ class Arnoldifyer(object):
         if full:
             Vh = numpy.c_[deflated_solver.V[:, :n],
                           deflated_solver.projection.U].dot(Wto.dot(QT))
-            F = - self.Z.dot(Rh.dot(Vh.T.conj()))
-            F = F + F.T.conj()
+            ip_Minv_B = deflated_solver.linear_system.get_ip_Minv_B()
+
+            def _apply_F(x):
+                '''Application of the perturbation.'''
+                return -(self.Z.dot(Rh.dot(utils.inner(Vh, x,
+                                                       ip_B=ip_Minv_B)))
+                         + Vh.dot(Rh.T.conj().dot(utils.inner(self.Z, x,
+                                                              ip_B=ip_Minv_B)))
+                         )
+            F = utils.LinearOperator((Vh.shape[0], Vh.shape[0]),
+                                     dtype=deflated_solver.dtype,
+                                     dot=_apply_F
+                                     )
             return Hh, Rh, q_norm, vdiff_norm, PWAW_norm, Vh, F
         return Hh, Rh, q_norm, vdiff_norm, PWAW_norm
 
 
-def bound_pseudo(arnoldifyer, Wt, b_norm,
+def bound_pseudo(arnoldifyer, Wt,
                  g_norm=0.,
                  G_norm=0.,
                  GW_norm=0.,
                  WGW_norm=0.,
                  tol=1e-6,
-                 method='gmres',
                  pseudo_type='auto',
+                 pseudo_kwargs=None,
                  delta_n=20,
-                 delta_exp=5
+                 terminate_factor=1.
                  ):
     r'''Bound residual norms of next deflated system.
 
@@ -384,20 +470,13 @@ def bound_pseudo(arnoldifyer, Wt, b_norm,
       :math:`W=[V,U]\tilde{W}` and
       :math:`\mathcal{W}=\operatorname{colspan}(W)`. Must fulfill
       :math:`\tilde{W}^*\tilde{W}=I_k`.
-    :param b_norm: norm :math:`\|b\|` of right hand side.
     :param g_norm: norm :math:`\|g\|` of difference :math:`g=c-b` of
-      right hand sides.
+      right hand sides. Has to fulfill :math:`\|g\|<\|b\|`.
     :param G_norm: norm :math:`\|G\|` of difference
       :math:`G=B-A` of operators.
     :param GW_norm: Norm :math:`\|G|_{\mathcal{W}}\|` of difference
       :math:`G=B-A` of operators restricted to :math:`\mathcal{W}`.
     :param WGW_norm: Norm :math:`\|\langle W,GW\rangle\|_2`.
-    :param method: type of method that is used to construct the residuals and
-      residual/error polynomial from :math:`\hat{H}`:
-
-      * ``'cg'``: for the CG method
-      * ``'minres'``: for the MINRES method
-      * ``'gmres'``: for the GMRES method
     :param pseudo_type: One of
 
       * ``'auto'``: determines if :math:`\hat{H}` is non-normal, normal or
@@ -411,21 +490,75 @@ def bound_pseudo(arnoldifyer, Wt, b_norm,
       * ``'contain'``: the pseudospectrum of the extended Hessenberg matrix
         :math:`\begin{bmatrix}\hat{H}\\S_i\end{bmatrix}` is used
         (pseudospectrum has to be re computed for each iteration).
+      * ``'omit'``: do not compute the pseudospectrum at all and just use the
+        residual bounds from the approximate Krylov subspace.
+    :param pseudo_kwargs: (optional) arguments that are passed to the method
+      that computes the pseudospectrum.
+    :param terminate_factor: (optional) terminate the computation if the ratio
+      of two subsequent residual norms is larger than the provided factor.
+      Defaults to 1.
     '''
+    if pseudo_kwargs is None:
+        pseudo_kwargs = {}
+
+    # Arnoldify!
     Hh, Rh, q_norm, vdiff_norm, PWAW_norm = arnoldifyer.get(Wt)
 
-    # smallest singular value of W^*AW
-    WAW = Wt.T.conj().dot(arnoldifyer.J.dot(arnoldifyer.L.dot(Wt)))
-    sigma_min = numpy.min(scipy.linalg.svdvals(WAW))
+    # get original linear system
+    ls_orig = arnoldifyer._deflated_solver.linear_system
 
-    if sigma_min <= WGW_norm:
-        raise utils.AssumptionError(
-            'sigma_min(W^*AW) > ||W^*GW|| not satisfied.')
-    eta = GW_norm/(sigma_min - WGW_norm)
+    k = Wt.shape[1]
+    if k > 0:
+        # smallest singular value of W^*AW
+        WAW = Wt.T.conj().dot(arnoldifyer.J.dot(arnoldifyer.L.dot(Wt)))
+        sigma_min = numpy.min(scipy.linalg.svdvals(WAW))
+
+        if sigma_min <= WGW_norm:
+            raise utils.AssumptionError(
+                'sigma_min(W^*AW) > ||W^*GW|| not satisfied.')
+        eta = GW_norm/(sigma_min - WGW_norm)
+    else:
+        eta = 0.
+    b_norm = ls_orig.MMlb_norm
     beta = PWAW_norm*(eta*(b_norm + g_norm) + g_norm) + vdiff_norm
 
+    # check assumption on g_norm and b_norm
+    if g_norm >= b_norm:
+        raise utils.AssumptionError(
+            '||g_norm|| < ||b_norm|| not satisfied')
+
+    # compute residual norms of Hh*z=e_1*b_norm
+    ls_small = linsys.LinearSystem(Hh,
+                                   numpy.eye(Hh.shape[0], 1) * q_norm,
+                                   normal=ls_orig.normal,
+                                   self_adjoint=ls_orig.self_adjoint,
+                                   positive_definite=ls_orig.positive_definite
+                                   )
+
+    Solver = type(arnoldifyer._deflated_solver)
+    if issubclass(Solver, linsys.Minres) or issubclass(Solver, linsys.Gmres):
+        aresnorms = utils.get_residual_norms(Hh,
+                                             self_adjoint=ls_orig.self_adjoint)
+    else:
+        # TODO: compute residuals more efficiently for CG
+        try:
+            solver = Solver(ls_small, tol=tol, maxiter=Hh.shape[0])
+        except utils.ConvergenceError as e:
+            # use all residuals that have been computed
+            # (useful for short recurrences)
+            solver = e.solver
+        aresnorms = numpy.array(solver.resnorms)
+    # absolute residual norm
+    aresnorms = aresnorms * q_norm
+
+    if pseudo_type == 'omit':
+        return aresnorms / (b_norm - g_norm)
+
     # spectrum of Hh
     evals, evecs = scipy.linalg.eig(Hh)
+    if ls_small.self_adjoint:
+        evals = numpy.real(evals)
+
     # norm of Hh
     Hh_norm = numpy.linalg.norm(Hh, 2)
 
@@ -444,66 +577,134 @@ def bound_pseudo(arnoldifyer, Wt, b_norm,
     if pseudo_type == 'auto':
         pseudo_type = _auto()
 
-    # compute residual norms of Hh*z=e_1*b_norm
-    if method == 'cg':
-        Solver = linsys.Cg
-    elif method == 'minres':
-        Solver = linsys.Minres
-    elif method == 'gmres':
-        Solver = linsys.Gmres
-    else:
-        raise utils.ArgumentError('unknown method: {0}'.format(method))
-    solver = Solver(Hh, numpy.eye(Hh.shape[0], 1)*b_norm,
-                    tol=tol, maxiter=Hh.shape[0])
+    # for delta >= min(|\lambda|), the pseudospectrum will contain zero and
+    # the thus polymax > 1. nevertheless, the bound may provide useful
+    # information in early iterations with large values of delta.
+    # Therefore, the maximal perturbation is chosen as the maximal
+    # eigenvalue of Hh
+    delta_max = 1e2*numpy.max(numpy.abs(evals))
+
+    # minimal delta is defined via Rh
+    # HACK until numpy.linal.svd (and thus numpy.linalg.norm) is fixed
+    from scipy.linalg import svd
+    _, Rhsvd, _ = svd(Rh[:, :1])
+    delta_min = PWAW_norm*(eta*(Hh_norm + G_norm) + G_norm) + numpy.max(Rhsvd)
+    if delta_min == 0:
+        delta_min = 1e-16
 
     import pseudopy
-    if pseudo_type == 'nonnormal':
-        # todo: construct bounding box!
-        raise NotImplementedError('nonnormal not yet implemented')
-    elif pseudo_type == 'normal':
+    if not ls_small.normal:
+        # construct pseudospectrum for the expected range
+        pseudo = pseudopy.NonnormalAuto(Hh, delta_min*0.99, delta_max*1.01,
+                                        **pseudo_kwargs)
+    elif not ls_small.self_adjoint:
         pseudo = pseudopy.NormalEvals(evals)
-    elif pseudo_type == 'hermitian':
-        raise NotImplementedError('hermitian not yet implemented')
+    else:
+        pseudo = None
 
-    bounds = []
-    for i in range(1, len(solver.resnorms)):
-        res = solver.resnorms[i]*b_norm
+    bounds = [aresnorms[0]]
+    for i in range(1, len(aresnorms)):
+        # compute roots of polynomial
+        if issubclass(Solver, linsys.Cg):
+            roots = scipy.linalg.eigvalsh(Hh[:i, :i])
+        else:
+            # TODO: more stable way of computing the roots of the MINRES
+            #       poly with exploitation of symmetry?
+            HhQ, HhR = scipy.linalg.qr(Hh[:i+1, :i], mode='economic')
+            roots_inv = scipy.linalg.eigvals(HhQ[:i, :].T.conj(), HhR)
+            roots = 1./roots_inv[numpy.abs(roots_inv) > 1e-14]
+
+        if ls_small.self_adjoint:
+            roots = numpy.real(roots)
+
+        # compute polynomial
+        p = utils.NormalizedRootsPolynomial(roots)
+        if ls_small.self_adjoint:
+            p_minmax_candidates = p.minmax_candidates()
+
+        # absolute residual
+        aresnorm = aresnorms[i]
+
+        # perturbation
+        # HACK until numpy.linal.svd (and thus numpy.linalg.norm) is fixed
+        from scipy.linalg import svd
+        _, Rhsvd, _ = svd(Rh[:, :i])
+        Rhnrm = numpy.max(Rhsvd)
         epsilon = PWAW_norm*(eta*(Hh_norm + G_norm) + G_norm) \
-            + numpy.linalg.norm(Rh[:, :i], 2)
+            + Rhnrm
+            #+ numpy.linalg.norm(Rh[:, :i], 2)
+        if epsilon == 0:
+            epsilon = 1e-16
+
         if pseudo_type == 'contain':
             raise NotImplementedError('contain not yet implemented')
-        pseudo_terms = []
-        # TODO: compute polynomial from (harmonic) Ritz values
-        for delta in epsilon*numpy.logspace(0, delta_exp, delta_n)[1:]:
-            # get pseudospectrum paths
-            pseudo_path = pseudo.contour_paths(delta)
 
-            # compute roots of polynomial
-            if method == 'cg':
-                roots = scipy.linalg.eigvalsh(Hh[:i, :i])
+        # exit if epsilon >= delta_max
+        if epsilon >= delta_max:
+            break
+
+        delta_log_range = numpy.linspace(numpy.log10(1.01*epsilon),
+                                         numpy.log10(delta_max),
+                                         delta_n+2
+                                         )[0:-1]
+
+        def compute_pseudo(delta_log):
+            delta = 10**delta_log
+            if ls_small.self_adjoint:
+                # pseudospectrum are intervals
+                pseudo_intervals = utils.Intervals(
+                    [utils.Interval(ev-delta, ev+delta) for ev in evals])
+
+                # add roots of first derivative of p
+                candidates = []
+                for candidate in p_minmax_candidates:
+                    if pseudo_intervals.contains(candidate):
+                        candidates.append(candidate)
+                all_candidates = numpy.r_[pseudo_intervals.get_endpoints(),
+                                          numpy.array(candidates)]
+
+                # evaluate polynomial
+                polymax = numpy.max(numpy.abs(p(all_candidates)))
+                pseudolen = 2 * delta
+
             else:
-                # TODO: more stable way of computing the roots of the MINRES
-                #       poly with exploitation of symmetry?
-                HhQ, HhR = scipy.linalg.qr(Hh[:i+1, :i], mode='economic')
-                roots_inv = scipy.linalg.eigvals(HhQ[:i, :].T.conj(), HhR)
-                roots = 1./roots_inv[numpy.abs(roots_inv) > 1e-14]
+                # get pseudospectrum paths
+                pseudo_path = pseudo.contour_paths(delta)
 
-            # compute polynomial
-            from numpy.polynomial import Polynomial as P
-            p = P.fromroots(roots)
+                # length of boundary
+                pseudolen = pseudo_path.length()
 
-            # evaluate polynomial on points of path
-            polymax = numpy.max(numpy.abs(p(pseudo_path.vertices())))
+                # evaluate polynomial on points of path
+                if pseudolen > 0:
+                    polymax = numpy.max(numpy.abs(p(pseudo_path.vertices())))
+                else:
+                    polymax = numpy.Inf
 
             # compute THE bound
-            pseudo_terms.append(
-                pseudo_path.length()/(2*numpy.pi*delta)
-                * (epsilon/(delta-epsilon)*(q_norm + beta) + beta)
+            return pseudolen/(2*numpy.pi*delta) \
+                * (epsilon/(delta-epsilon)*(q_norm + beta) + beta) \
                 * polymax
-                )
-        # append minimal vound value
-        bounds.append(res + numpy.min(pseudo_terms))
-    return numpy.array(bounds)
+
+        # minimization
+        from scipy.optimize import minimize_scalar
+        opt_res = minimize_scalar(compute_pseudo,
+                                  bounds=(delta_log_range[0],
+                                          delta_log_range[-1]),
+                                  method='bounded',
+                                  options={'maxiter': delta_n}
+                                  )
+        # the delta with minimal value is min_delta = 10**opt_res.x
+        min_val = opt_res.fun
+
+        # minimal bound value
+        boundval = aresnorm + min_val
+
+        # if not increasing: append to bounds
+        if i > 1 and boundval/bounds[-1] > terminate_factor:
+            break
+        else:
+            bounds.append(numpy.min([boundval, bounds[-1]]))
+    return numpy.array(bounds) / (b_norm - g_norm)
 
 
 class Ritz(object):
@@ -552,9 +753,7 @@ class Ritz(object):
             # build block matrices
             M = numpy.bmat([[H + B.dot(EinvC), B],
                             [C, E]])
-            # TODO: has M to be applied in the next line on one side?
-            # Andre: I think so! :)
-            F = utils.inner(projection.AU, projection.AU,
+            F = utils.inner(projection.AU, projection.MAU,
                             ip_B=linear_system.ip_B)
             S = numpy.bmat([[I(n_), B_, O((n_, m))],
                             [B_.T.conj(), F, E],
