@@ -12,6 +12,7 @@ import scipy.linalg
 from scipy.sparse import isspmatrix
 #from scipy.sparse.linalg import LinearOperator, aslinearoperator
 from scipy.sparse.sputils import isintlike
+from collections import defaultdict
 
 # for Givens rotations
 try:
@@ -23,7 +24,7 @@ except ImportError:
     import scipy.linalg.blas as blas
 
 __all__ = ['ArgumentError', 'AssumptionError', 'ConvergenceError',
-           'LinearOperatorError', 'InnerProductError',
+           'LinearOperatorError', 'InnerProductError', 'RuntimeError',
            'Arnoldi', 'BoundCG', 'BoundMinres',
            'ConvergenceError', 'Givens', 'House', 'IdentityLinearOperator',
            'LinearOperator', 'MatrixLinearOperator',
@@ -73,6 +74,10 @@ class InnerProductError(Exception):
     '''Raised when the inner product is indefinite.'''
 
 
+class RuntimeError(Exception):
+    '''Raised for errors that do not fit in any other exception.'''
+
+
 def find_common_dtype(*args):
     '''Returns common dtype of numpy and scipy objects.
 
@@ -196,14 +201,16 @@ def norm(x, y=None, ip_B=None):
                       or isinstance(ip_B, IdentityLinearOperator)):
         return numpy.linalg.norm(x, 2)
     ip = inner(x, y, ip_B=ip_B)
-    nrm = numpy.sqrt(numpy.linalg.norm(ip, 2))
-    if numpy.linalg.norm(ip.imag, 2) > nrm*1e-10:
+    nrm_diag = numpy.linalg.norm(numpy.diag(ip), 2)
+    nrm_diag_imag = numpy.linalg.norm(numpy.imag(numpy.diag(ip)), 2)
+    if nrm_diag_imag > nrm_diag*1e-10:
         raise InnerProductError('inner product defined by ip_B not positive '
-                                'definite?')
-    return nrm
+                                'definite? ||diag(ip).imag||/||diag(ip)||={0}'
+                                .format(nrm_diag_imag/nrm_diag))
+    return numpy.sqrt(numpy.linalg.norm(ip, 2))
 
 
-def get_linearoperator(shape, A):
+def get_linearoperator(shape, A, timer=None):
     """Enhances aslinearoperator if A is None."""
     ret = None
     import scipy.sparse.linalg as scipylinalg
@@ -222,8 +229,16 @@ def get_linearoperator(shape, A):
                              dtype=A.dtype)
     else:
         raise TypeError('type not understood')
+
+    # set up timer if requested
+    if A is not None and not isinstance(A, IdentityLinearOperator) \
+            and timer is not None:
+        ret = TimedLinearOperator(ret, timer)
+
+    # check shape
     if shape != ret.shape:
         raise LinearOperatorError('shape mismatch')
+
     return ret
 
 
@@ -356,27 +371,97 @@ class House:
         return numpy.eye(n, n) - self.beta * numpy.dot(self.v, self.v.T.conj())
 
 
-def hessenberg(A):
-    '''Hessenberg decomposition via Householder transformations.
+def hessenberg(a, calc_q=False, overwrite_a=False, check_finite=True):
+    """
+    Compute Hessenberg form of a matrix.
 
-    The function ``scipy.linalg.hessenberg`` suffers from a bug, see
-    https://github.com/scipy/scipy/issues/3364.
+    The Hessenberg decomposition is::
 
-    This function implements algorithm 7.4.2 in [GolV13]_.
+        A = Q H Q^H
 
-    :returns: ``H`` and ``Q`` such that ``A = Q H Q^*``
-    '''
-    n = A.shape[0]
-    if A.shape[1] != n:
-        raise ArgumentError('matrix must be square')
-    H = A.copy()
-    Q = numpy.eye(n, dtype=H.dtype)
-    for k in range(n-2):
-        house = House(H[k+1:, [k]])
-        H[k+1:, k:] = house.apply(H[k+1:, k:])
-        H[:, k+1:] = house.apply(H[:, k+1:].T.conj()).T.conj()
-        Q[k+1:, :] = house.apply(Q[k+1:, :])
-    return H, Q.T.conj()
+    where `Q` is unitary/orthogonal and `H` has only zero elements below
+    the first sub-diagonal.
+
+    Parameters
+    ----------
+    a : (M, M) array_like
+        Matrix to bring into Hessenberg form.
+    calc_q : bool, optional
+        Whether to compute the transformation matrix.  Default is False.
+    overwrite_a : bool, optional
+        Whether to overwrite `a`; may improve performance.
+        Default is False.
+    check_finite : boolean, optional
+        Whether to check that the input matrix contains only finite numbers.
+        Disabling may give a performance gain, but may result in problems
+        (crashes, non-termination) if the inputs do contain infinities or NaNs.
+
+    Returns
+    -------
+    H : (M, M) ndarray
+        Hessenberg form of `a`.
+    Q : (M, M) ndarray
+        Unitary/orthogonal similarity transformation matrix ``A = Q H Q^H``.
+        Only returned if ``calc_q=True``.
+
+    """
+    from numpy import asarray_chkfinite, asarray
+    from scipy.linalg import calc_lwork
+    from scipy.linalg.misc import _datacopied
+    from scipy.linalg.lapack import get_lapack_funcs
+    from scipy.linalg.blas import get_blas_funcs
+
+    if check_finite:
+        a1 = asarray_chkfinite(a)
+    else:
+        a1 = asarray(a)
+    if len(a1.shape) != 2 or (a1.shape[0] != a1.shape[1]):
+        raise ValueError('expected square matrix')
+    overwrite_a = overwrite_a or (_datacopied(a1, a))
+
+    # if 2x2 or smaller: already in Hessenberg form
+    if a1.shape[0] <= 2:
+        if calc_q:
+            return a1, numpy.eye(a1.shape[0])
+        return a1
+
+    gehrd, gebal = get_lapack_funcs(('gehrd', 'gebal'), (a1,))
+    ba, lo, hi, pivscale, info = gebal(a1, permute=0, overwrite_a=overwrite_a)
+    if info < 0:
+        raise ValueError('illegal value in %d-th argument of internal gebal '
+                         '(hessenberg)' % -info)
+    n = len(a1)
+    lwork = calc_lwork.gehrd(gehrd.typecode, n, lo, hi)
+    hq, tau, info = gehrd(ba, lo=lo, hi=hi, lwork=lwork, overwrite_a=1)
+    if info < 0:
+        raise ValueError('illegal value in %d-th argument of internal gehrd '
+                         '(hessenberg)' % -info)
+
+    if not calc_q:
+        for i in range(lo, hi):
+            hq[i+2:hi+1, i] = 0.0
+        return hq
+
+    # XXX: Use ORGHR routines to compute q.
+    typecode = hq.dtype
+    ger, gemm = get_blas_funcs(('ger', 'gemm'), dtype=typecode)
+    q = None
+    for i in range(lo, hi):
+        if tau[i] == 0.0:
+            continue
+        v = numpy.zeros(n, dtype=typecode)
+        v[i+1] = 1.0
+        v[i+2:hi+1] = hq[i+2:hi+1, i]
+        hq[i+2:hi+1, i] = 0.0
+        h = ger(-tau[i], v, v,
+                a=numpy.diag(numpy.ones(n, dtype=typecode)), overwrite_a=1)
+        if q is None:
+            q = h
+        else:
+            q = gemm(1.0, q, h)
+    if q is None:
+        q = numpy.diag(numpy.ones(n, dtype=typecode))
+    return hq, q
 
 
 class Givens:
@@ -417,6 +502,7 @@ class Projection(object):
     def __init__(self, X,
                  Y=None,
                  ip_B=None,
+                 orthogonalize=True,
                  iterations=2
                  ):
         """Generic projection.
@@ -446,6 +532,10 @@ class Projection(object):
         :param ip_B: (optional) inner product, see :py:meth:`inner`. ``None``,
             a ``numpy.array`` or a ``LinearOperator`` is preferred due to the
             applicability of the proposed algorithms in [Ste11]_, see below.
+        :param orthogonalize: (optional) `True` orthogonalizes the bases
+            provided in `X` and `Y` with respect to the inner product defined
+            by `ip_B`. Defaults to `True` as the orthogonalization is
+            suggested by the round-off error analysis in [Ste11]_.
         :param iterations: (optional) number of applications of the projection.
             It was suggested in [Ste11]_ to use 2 iterations (default) in order
             to achieve high accuracy ("twice is enough" as in the orthogonal
@@ -460,6 +550,7 @@ class Projection(object):
         self.ip_B = ip_B
         if iterations < 1:
             raise ArgumentError('iterations < 1 not allowed')
+        self.orthogonalize = orthogonalize
         self.iterations = iterations
 
         Y = X if Y is None else Y   # default: orthogonal projection
@@ -477,13 +568,21 @@ class Projection(object):
             return
 
         # orthogonalize X
-        self.V, self.VR = qr(X, ip_B=ip_B)
+        if orthogonalize:
+            self.V, self.VR = qr(X, ip_B=ip_B)
+        else:
+            self.V = X
+            self.VR = None
 
-        if Y is X:  # orthogonal projection
+        if Y is X and orthogonalize:  # orthogonal projection
             self.W, self.WR = self.V, self.VR
             self.Q, self.R = None, None
         else:  # general case
-            self.W, self.WR = qr(Y, ip_B=ip_B)
+            if orthogonalize:
+                self.W, self.WR = qr(Y, ip_B=ip_B)
+            else:
+                self.W = Y
+                self.WR = None
             M = inner(self.W, self.V, ip_B=ip_B)
             self.Q, self.R = scipy.linalg.qr(M)
 
@@ -877,7 +976,7 @@ class Arnoldi(object):
         self.invariant = False
 
         if ortho == 'house':
-            if (M is not None
+            if (self.M is not None
                     and not isinstance(self.M, IdentityLinearOperator)) or \
                     (not isinstance(self.ip_B, IdentityLinearOperator) and
                      self.ip_B is not None):
@@ -1246,7 +1345,7 @@ def ritz(H, V=None, hermitian=False, type='ritz'):
     return theta, U, resnorm
 
 
-class Timer:
+class Timer(list):
     """Measure execution time of multiple code blocks with ``with``.
 
     Example: ::
@@ -1266,12 +1365,23 @@ class Timer:
         time me, too!
         [6.389617919921875e-05, 6.008148193359375e-05]
 
-    If you want to measure different types of code blocks you can use
-    ``defaultdict`` as a timer manager: ::
+    """
+    def __init__(self):
+        super(Timer, self).__init__()
 
-        from collections import defaultdict
+    def __enter__(self):
+        self.tstart = time.time()
 
-        tm = defaultdict(Timer)
+    def __exit__(self, a, b, c):
+        self.append(time.time() - self.tstart)
+
+
+class Timings(defaultdict):
+    '''Manages several timers.
+
+    If you want to measure different types of code blocks you can use ::
+
+        tm = Timings()
         with tm['class1']:
             print('code that belongs to class1')
         with tm['class2']:
@@ -1286,18 +1396,29 @@ class Timer:
         code that belongs to class2
         code that belongs to class1
         defaultdict(<class krypy.utils.Timer at 0x23810b8>, {'class2': [2.9087066650390625e-05], 'class1': [4.696846008300781e-05, 3.2901763916015625e-05]})
-    """
+    '''
     def __init__(self):
-        self.times = []
+        super(Timings, self).__init__(Timer)
 
-    def __enter__(self):
-        self.tstart = time.time()
+    def get(self, key):
+        '''Return timings for `key`. Returns 0 if not present.'''
+        if key in self and len(self[key]) > 0:
+            return min(self[key])
+        else:
+            return 0
 
-    def __exit__(self, a, b, c):
-        self.times.append(time.time() - self.tstart)
+    def get_ops(self, ops):
+        '''Return timings for dictionary ops holding the operation names as
+        keys and the number of applications as values.'''
+        time = 0.
+        for op, count in ops.iteritems():
+            time += self.get(op) * count
+        return time
 
     def __repr__(self):
-        return self.times.__repr__()
+        return 'Timings(' + ', '.join(
+            ['{0}: {1}'.format(key, self.get(key)) for key in self]
+            ) + ')'
 
 
 class LinearOperator(object):
@@ -1534,6 +1655,40 @@ class MatrixLinearOperator(LinearOperator):
         return self._A.__repr__()
 
 
+class TimedLinearOperator(LinearOperator):
+    def __init__(self, linear_operator, timer=None):
+        self._linear_operator = linear_operator
+        super(TimedLinearOperator, self).__init__(
+            shape=linear_operator.shape,
+            dtype=linear_operator.dtype,
+            dot=linear_operator.dot,
+            dot_adj=linear_operator.dot_adj
+            )
+
+        if timer is None:
+            timer = Timer()
+        self._timer = timer
+
+    def dot(self, X):
+        k = X.shape[1]
+        if k == 0:
+            return self._linear_operator.dot(X)
+        with self._timer:
+            ret = self._linear_operator.dot(X)
+        self._timer[-1] /= k
+        return ret
+
+    def dot_adj(self, X):
+        k = X.shape[1]
+        if k == 0:
+            return self._linear_operator.dot(X)
+        k = X.shape[1]
+        with self._timer:
+            ret = self._linear_operator.dot_adj(X)
+        self._timer[-1] /= k
+        return ret
+
+
 def strakos(n, l_min=0.1, l_max=100, rho=0.9):
     """Return the Strakoš matrix.
 
@@ -1768,6 +1923,10 @@ class BoundCG(object):
                 raise AssumptionError(
                     'non-positive eigenvalues not allowed with intervals')
 
+        # empty spectrum?
+        if len(evals) == 0:
+            raise AssumptionError('empty spectrum not allowed')
+
         # all evals real?
         if not numpy.isreal(evals).all():
             raise AssumptionError('non-real eigenvalues not allowed')
@@ -1795,7 +1954,7 @@ class BoundCG(object):
 
     def get_step(self, tol):
         '''Return step at which bound falls below tolerance.'''
-        return int(numpy.ceil(numpy.log(tol/2.)/numpy.log(self.base)))
+        return numpy.log(tol/2.)/numpy.log(self.base)
 
 
 class BoundMinres(object):
@@ -1848,6 +2007,10 @@ class BoundMinres(object):
                                      evals.min_pos(), evals.max()]
                      if val is not None]
 
+        # empty spectrum?
+        if len(evals) == 0:
+            raise AssumptionError('empty spectrum not allowed')
+
         # all evals real?
         if not numpy.isreal(evals).all():
             raise AssumptionError('non-real eigenvalues not allowed')
@@ -1876,7 +2039,7 @@ class BoundMinres(object):
 
     def get_step(self, tol):
         '''Return step at which bound falls below tolerance. '''
-        return int(2 * numpy.ceil(numpy.log(tol/2.)/numpy.log(self.base)))
+        return 2 * numpy.log(tol/2.)/numpy.log(self.base)
 
 
 def bound_perturbed_gmres(pseudo, p, epsilon, deltas):
@@ -1968,3 +2131,24 @@ class NormalizedRootsPolynomial(object):
         if numpy.isscalar(points):
             return numpy.asscalar(vals)
         return vals
+
+
+def get_residual_norms(H, self_adjoint=False):
+    '''Compute relative residual norms from Hessenberg matrix.
+
+    It is assumed that the initial guess is chosen as zero.'''
+    H = H.copy()
+    n_, n = H.shape
+    y = numpy.eye(n_, 1, dtype=H.dtype)
+    resnorms = [1.]
+    for i in range(n_-1):
+        G = Givens(H[i:i+2, [i]])
+        if self_adjoint:
+            H[i:i+2, i:i+3] = G.apply(H[i:i+2, i:i+3])
+        else:
+            H[i:i+2, i:] = G.apply(H[i:i+2, i:])
+        y[i:i+2] = G.apply(y[i:i+2])
+        resnorms.append(numpy.abs(y[i+1, 0]))
+    if n_ == n:
+        resnorms.append(0.)
+    return numpy.array(resnorms)
